/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
// ceDonkey - a free, simple (?), dirty eDonkey server
// 
// planned features (guaranteed):
//   - statistical analysis of things:
//       - connection times
//       - download behaviour 
//       - search terms
//       - client analysis (from connection name)
//       - file availability to file demand
//       - traffic statistics 
//       - client protocol behaviour 
//       - shared files per user statistics
//       - hash <-> file statistics 
//       - ...
//   - blacklists (IP network addresses)
//   - blacklists (user names)
//   - blacklists (connection behaviour - like bots or so)
//   - full protocol compatibility, except of additional
//     data in items (bitrate, frames/s, etc)
//   - ability to understand FQDN's as server address
//   - better handling of delivering search results, see forthcoming FAQ 
//   - UDP search request limiting
//
// Tech details:
//   - based on GNet / GLib
//   - telnet based control
// 
// Features NOT supported:
//   - groups (only if this is a server functionality :-)
//   - some else I don't quite remember
//   - plugin functionality (too advanced for now, this is a 
//     straightforward server)
//   - error control - be stupid, and it will crash
//
// All comments & suggestions to edonkey@the-me.de
// 
// ATTENTION: help is always welcome ... :-))


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gnet/gnet.h>
#include <signal.h>
#include <glib.h>

#include "structs.h"
#include "littlehelpers.h"
#include "tcp_connections.h"
#include "udp_handler.h"
#include "ed2k_message_handler.h"

#include "maintenance.h"
#include "timing.h"
#include "init.h"
#include "tcp_listeners.h"


extern struct central_struct *rome;



/* the only valid method to abort :-) */
static  void signal_catch_sigint(int signum)
{
   g_print("[ SIG ] SIGINT caught\n");

   exit(EXIT_FAILURE);
}





void create_threads()
{
   GError *error;
   gint    i,j;
   struct PollContext *P1, *P2;
   
   error = NULL;



   if (! rome->tcp_workers || ! rome->udp_workers)
      {
	 printf("[FATAL] you didn't set tcpworkers or udpworkers in config file!\n");
	 printf("        aborting.\n");
	 exit(-1);
      }
   printf("creating tcp workers\n");
   j = rome->tcp_workers;
   P2 = NULL;
   for (i=0; i<j; i++)
      {
	 P1 = g_new0(struct PollContext, 1);
	 P1->context = g_main_context_new();
	 P1->loop    = g_main_loop_new(P1->context, FALSE);
	 P1->thread  = g_thread_create((GThreadFunc) tcp_context_listener, 
				       P1, FALSE, &error);
	 if (error)
	    {
	       printf("[FATAL] failed to create tcp workers\n"); 
	       printf("    --> %s\n", error->message);
	       printf("        aborting.\n");
	       exit(-1);
	    }

	 P1->count   = i;

	 if (P2)
	    {
	       P1->next = P2->next;
	       P1->prev = P2;
	       P2->next = P1;
	       P2->next->prev = P1;
	    }
	 else
	    {
	       P1->next = P1;
	       P1->prev = P1;
	       P2 = P1;
	    }
      }
   rome->contexts = P1->next;

#ifdef debug
   P1 = P2->next;
   printf("checking workers: ");
   do {
      printf("%d ", P1->count);
      P1 = P1->next;
   } while (P1 != P2);
   printf("\n");
#endif



   /* init thread pool for packet processing */
   rome->udp_pool = g_thread_pool_new((GFunc)ed2k_message_popper, 
				      rome, 1, TRUE, &error);
   if (error)
      {
	 printf("[FATAL] error creating udp thread pool!\n");
	 printf("    --> %s\n", error->message);
	 printf("        aborting.\n");
	 exit(-1);
      }
   
   /* the thread handling udp connections */
   rome->udp_thread = g_thread_create((GThreadFunc) udp_message_handler, 
				       rome, TRUE, &error);
   if (error)
      {
	 printf("[FATAL] error creating tcp listeners!\n");
	 printf("    --> %s\n", error->message);
	 printf("        aborting.\n");
	 exit(-1);
      }   
}





void
perform_basic_test()
{
#ifndef G_THREADS_ENABLED
   printf("critical: threads support not enabled in GLib. Aborting.\n");
   exit(-1);
#endif

   if ( ! g_thread_supported() ) {
      g_thread_init(NULL);
      printf("[INFO] glib threads initialized\n");
   }
   else {
      printf("[ERROR] wups. this shouldn't happen - thread problem?\n");
      printf("        aborting!\n");
      exit(-1);
   }

}



int main(int argc, char **argv)
{
   /* the magic begins ... */

   perform_basic_test();

   rome = g_new0(struct central_struct, 1);
   printf("rome = %d\n", (unsigned int) rome);

   ed2k_init();

#ifdef debug
   printf("ceDonkey.c: rome = %s\n", rome->string);
#endif

   create_threads();

   tcp_setup();
   timing_setup();

   signal(SIGINT, signal_catch_sigint);


   // run the bastard :-))
   g_main_loop_run(rome->main_loop);

   
   return 0;

}
   


