/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
/* 
   file     : ed2k_data.h
   purpose  : header file for ed2k.c

*/


/* *********************************************************************** *
 * note:                                                                   *
 * *********************************************************************** *
 * this file is the general interface to the ed2k data structure. all      *
 * functions within here can be replaced without any worry, AS LONG AS     *
 * THEY KEEP BEHAVING TO THE OUTSIDE AS INTENDED. so the complete replace- *
 * ment of the backend database engine should (!) be no problem ... :-)    *
 *                                                                         *
 * so this should (!!!) be the complete api to the edonkey server system   *
 * *********************************************************************** */



#ifndef __ed2k_data
#define __ed2k_data 1

#include "structs.h"


/* *********************************************************************** *
 * the init - call first and once                                          *
 * *********************************************************************** */

int ed2k_data_init();




/* *********************************************************************** *
 * counting, statistical and helper stuff                                  *
 * *********************************************************************** *
 * these may be called from anywhere                                       *
 * *********************************************************************** */


int ed2k_data_get_all_files_count();
int ed2k_data_get_different_files_count();
int ed2k_data_get_users_count();
int ed2k_data_get_lowid_count();

int ed2k_data_count_different_files();
int ed2k_data_count_users();

gint ed2k_data_get_files_byte_count();

// free the list returned by search_file_desc function - to be obsolete??
int ed2k_data_free_search_results(struct SearchResults *result);


// free the message_handler_struct struct
int ed2k_data_free_message(struct message_handle_struct *);



/* *********************************************************************** *
 * more serious stuff, best called from as few places as possible          *
 * *********************************************************************** */


void
ed2k_data_read_lock();

void
ed2k_data_read_unlock();

void
ed2k_data_write_lock();

void
ed2k_data_write_unlock();


int 
ed2k_data_user_delete(struct UserInfo *user);


struct SearchResults *
ed2k_data_search_file_desc(char **words);


int 
ed2k_data_user_new(struct UserInfo *user, 
		   gchar   *user_hash,
		   gchar   *user_name,
		   int      name_length,
		   gboolean open);


struct FileInfo *
ed2k_data_file_new(char *hash, int size, 
		   struct UserInfo *owner, 
		   char *name, unsigned short name_length);

struct FileInfo *
ed2k_data_file_add(char *hash, int size, 
		   struct UserInfo *owner, 
		   char *name, unsigned short name_length);


struct DescInfo *
ed2k_data_file_add_owner(struct FileInfo *, struct UserInfo *, 
			 char *, unsigned short);


struct SearchResults *
ed2k_data_generate_dl_sources(struct FileInfo *);


struct UserInfo *
ed2k_data_get_user_from_id(guint id);


struct FileInfo *
ed2k_data_search_file_hash(char *);




/* *********************************************************************** *
 * debug things. dont use ...                                              *
 * *********************************************************************** */

void debug_dump();




#endif



