/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#include <gnet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>

#include "structs.h"
#include "littlehelpers.h"
#include "ed2k_data.h"
#include "tcp_connections.h"
#include "serverlist.h"
#include "init.h"
#include "logging.h"
#include "settings.h"


/* ************************************************************************** *
                                the globals                              
   ************************************************************************** */

extern struct central_struct   *rome;

GRand                          *chaos;




/* ************************************************************************** *
                               UDP requesting
   ************************************************************************** */


// 0x96
gint
ed2k_send_udp_ping(GInetAddr *addr)
{
   GUdpPacket udp_packet;
   guchar packet[6];
   
   packet[1] = 0xe3;
   packet[1] = 0x96;
   *((int*)(packet+2)) = g_rand_int(chaos);
   
   udp_packet.data   = (guchar*) &packet;
   udp_packet.length = 6;
   udp_packet.addr   = addr;

   count_udp_traffic(U96, udp_packet.length, UDP_OUT);
   log_write_0x96(addr, *(guint*)(packet+2), UDP_OUT);

   server_register_ping(addr, *((int*)(packet+2)));

   gnet_inetaddr_set_port(addr, gnet_inetaddr_get_port(addr)+4);
   gnet_udp_socket_send(rome->udp_socket, &udp_packet);
   gnet_inetaddr_set_port(addr, gnet_inetaddr_get_port(addr)-4);
   
   
   return 1;
}


// 0xa0 - server announce
gint
ed2k_send_udp_srv_announce(GInetAddr *addr)
{
   GUdpPacket udp_packet;
   guchar packet[8];
   

   packet[0] = 0xe3;
   packet[1] = 0xa0;

   *((guint*)(packet+2))   = rome->pure_ipaddr;
   *((gushort*)(packet+6)) = rome->pure_conn_port;
   
   udp_packet.data   = (guchar*) &packet;
   udp_packet.length = 8;
   udp_packet.addr   = addr;

   count_udp_traffic(Ua0, udp_packet.length, UDP_OUT);
   log_write_0xa0(addr, NULL, UDP_OUT);


   gnet_inetaddr_set_port(addr, gnet_inetaddr_get_port(addr)+4);
   gnet_udp_socket_send(rome->udp_socket, &udp_packet);
   gnet_inetaddr_set_port(addr, gnet_inetaddr_get_port(addr)-4);

   
   return 1;
}




// 0xa4 ??
gint
ed2k_send_udp_req_srvlist(GInetAddr *addr)
{
   GUdpPacket udp_packet;
   guchar packet[2];
   
   packet[0] = 0xe3;
   packet[1] = 0xa4;

   udp_packet.data   = (guchar*) &packet;
   udp_packet.length = 2;
   udp_packet.addr   = addr;

   count_udp_traffic(Ua4, udp_packet.length, UDP_OUT);
   log_write_0xa4(addr, UDP_OUT);

   gnet_inetaddr_set_port(addr, gnet_inetaddr_get_port(addr)+4);
   gnet_udp_socket_send(rome->udp_socket, &udp_packet);
   gnet_inetaddr_set_port(addr, gnet_inetaddr_get_port(addr)-4);

   
   return 1;
}




// TODO: IMPLEMENT
gint
ed2k_send_udp_push_fail(GInetAddr *addr)
{

   return 1;
}





/* ************************************************************************** *
                               UDP answering
   ************************************************************************** */




// 0x97
gint 
ed2k_send_udp_pong(struct message_handle_struct *message)
{
   
   guchar      *buffer;
   gint         users;
   gint         files;
 
   // free size always 4096
   buffer = message->buffer;
   
   memset(buffer+1, 0x97, 1);

   users = ed2k_data_get_users_count();
   memcpy(buffer+6 , &users, 4);
   files = ed2k_data_get_all_files_count();
   memcpy(buffer+10, &files, 4);
   
   count_udp_traffic(U97, 14, UDP_OUT);
   log_write_0x97(message->udp_packet->addr, *(guint*)(buffer+2), UDP_OUT);

   message->udp_packet->length = 14;
   gnet_udp_socket_send(rome->udp_socket, message->udp_packet);

   return 1;
}




// 0x9b
gint
ed2k_send_udp_download_sources(GInetAddr            *address, 
			       char                 *filehash,
			       struct SearchResults *sources,
			       int                   keeplock)
{
   GUdpPacket packet;
   char buffer[1600];
   gint act_pos;
   gint num_res;
   gint i;
   struct SearchResults *backup;

   // FUTURE: implement traffic control thing here (and all other udp funcs)

   
   backup = sources;
   num_res = 0;
   while (backup)
      {
	 num_res++;
	 backup = backup->next;
      }

   if (num_res > MAX_RESULTS_TO_SEND)
      num_res = MAX_RESULTS_TO_SEND;

   buffer[0] = 0xe3;
   buffer[1] = 0x9b;

   memcpy(buffer+2, filehash, 16);
   
   buffer[18] = (char) num_res;

   act_pos = 19;
   for (i = 0; i < num_res; i++)
      {
	 *((int*)    (buffer+act_pos  )) = sources->user->id;
	 *((gushort*)(buffer+act_pos+4)) = sources->user->listen_port;
	 act_pos += 6;
      }
   
   if (!keeplock)
      ed2k_data_read_unlock();
   
   packet.addr   = address;
   packet.length = (num_res * 6) + 19;
   packet.data   = (guchar*) &buffer;

   count_udp_traffic(U9b, packet.length, UDP_OUT);
   log_write_0x9b(address, num_res);

   gnet_udp_socket_send(rome->udp_socket, &packet);
   

   return num_res;
}



// TODO: test!
// 0xa1
gint 
ed2k_send_udp_serverlist(GInetAddr *address)
{
   struct ServerList *servers;
   struct ServerList *backup;

   GUdpPacket         packet;
   char               buffer[1600];
   gint               act_pos;
   gint               num_res;

   

   // FUTURE: implement traffic control thing here (and all other udp funcs)

   // format: own address, num of following, nof * (server address)
   
   server_lock();
   
   servers = server_get_all();

   if (!servers)
      {
	 server_unlock();
	 return 0;
      }

   buffer[0] = 0xe3;
   buffer[1] = 0xa1;
   
   // first our own address ... :-)
   memcpy(buffer+2, &(rome->pure_ipaddr), 6);

   act_pos = 9;
   num_res = 0;
   backup = servers;
   while (servers != backup && num_res < MAX_SERVERS_TO_SEND)
      {
	 memcpy(buffer+act_pos, &(servers->IP), 6);
	 servers = servers->next;
	 act_pos += 6;
	 num_res++;
      } 


   buffer[8] = (char) num_res;
   
   packet.addr   = address;
   packet.length = (num_res * 6) + 9;
   packet.data   = (guchar*) &buffer;

   count_udp_traffic(Ua1, packet.length, UDP_OUT);
   log_write_0xa1(address, num_res, UDP_OUT);

   gnet_udp_socket_send(rome->udp_socket, &packet);

   server_unlock();

   return num_res;
}


/* ************************************************************************** *
                                 TCP answering
   ************************************************************************** */



// 0x32
gint
ed2k_send_serverlist(struct UserInfo *user)
{
   guchar                  buffer[10000];
   guchar                 *pos;
   guint                   packet_size;
   struct ServerList      *list;
   struct ServerList      *backup;
   guint                   num_servers;
   guint                   written;

   buffer[0] = 0xe3;
   buffer[5] = 0x32;

   server_lock();

   list        = server_get_all();
   num_servers = 0;
   pos         = &(buffer[0]) + 7;
   backup      = list;

   if (list)
      while (list != backup && num_servers < MAX_SERVERS_TO_SEND)
	 {
	    *(int*)pos         = list->IP; 
	    *(gushort*)(pos+4) = list->port;
	    pos += 6;
	    
	    list = list->next;
	    num_servers++;
	 }

   packet_size = pos - &(buffer[0]);
   
   *(int*)(buffer+1)    = packet_size - 5;
   *(guchar*)(buffer+6) = num_servers;

   log_write_0x32(user);
   count_tcp_traffic(user, T32, packet_size - 5);

   g_io_channel_write_chars(user->iochannel, buffer,
			    packet_size, &written, NULL);

   g_assert(written==packet_size);
   
   g_io_channel_flush(user->iochannel, NULL);

   server_unlock();

   return 1;
}





// 0x33
// testing & verifying with protocol version 59
gint 
ed2k_return_search_results(struct UserInfo      *user,
			   struct SearchResults *results,
			   int                   keeplock)
{
   
   // calculate size of packet following:
   // size is:
   //     1         *   5 bytes (0x33 and num_results)
   //  +  num_files * \left
   //                    48 bytes (static data)
   //                  +  x bytes (length filename)
   //                  +  n bytes (length additional information) 
   //                 \right

   // additional information NOT taken care of right now ...


   struct SearchResults *backup;
   int packet_size, num_res;
   int t, written;
   char *pos;
   char buffer[10000];


   backup = results;

   buffer[0] = 0xe3;
   buffer[5] = 0x33;

   pos = &(buffer[10]);
   backup  = results;
   num_res = 0;
   while (backup)
      {
	 
	 // hash
	 memcpy(pos, &(backup->file->hash), 16);
	 pos += 16;

	 // some ip. nothing right now
	 *(int*)(pos  ) = 0;
	 *(int*)(pos+4) = 0;
	 pos += 6;

	 // tag count: 3 for now (name, size, avail)
	 *(int*)(pos) = 3;
	 pos += 4;

	 // tag: name
	 t = backup->desc->strlen;
	 *(int*)(pos)       = TAG_NAME_FOLLOWS;
	 *(gushort*)(pos+4) = t;
	 memcpy(pos+6, backup->desc->string, t);
	 pos += t + 6;
	 
	 // tag: size
	 *(int*)(pos)   = TAG_FILE_SIZE;
	 *(int*)(pos+4) = backup->file->size;
	 pos +=8;
	 
	 *(int*)(pos)   = TAG_FILE_AVAIL;
	 *(int*)(pos+4) = backup->file->count;
	 pos += 8;

	 backup = backup->next;
	 num_res++;
      }

   
   if (!keeplock)
      ed2k_data_read_unlock();
   

   *pos = 0;
   pos += 1;

   packet_size = pos - &(buffer[0]);

   *((int*)(buffer+6)) = num_res;
   *((int*)(buffer+1)) = packet_size - 5;

   count_tcp_traffic(user, T33, packet_size - 5);
   log_write_0x33(user, num_res);

   gnet_io_channel_writen(user->iochannel, buffer, packet_size, &written);
   
   g_io_channel_flush(user->iochannel, NULL);

   
   return 1;
}





// 0x34
gint
ed2k_send_server_status(struct UserInfo *user)
{
   guchar   header[14];
   guint    written;

   count_tcp_traffic(user, T34, 14);
   log_write_0x34(user);

   *((char*)(header   )) = 0xe3;
   *((int*) (header+1 )) = 9;
   *((int*) (header+5 )) = 0x34;
   *((int*) (header+6 )) = ed2k_data_get_users_count();
   *((int*) (header+10)) = ed2k_data_get_all_files_count();

   gnet_io_channel_writen(user->iochannel, header, 14, &written);
   g_assert(written == 14);


   return 1;
}




// 0x35
gint
ed2k_send_push_command(struct UserInfo *to_user, 
		       guint push_addr, gushort push_port)
{
   char buffer[12];

   buffer[0] = 0xe3;
   *((int*)(buffer+1)) = 7;
   buffer[5] = 0x35;

   // WHEW! +6 & +12 ?!? (old values)
   *((int*)(buffer+6))  = push_addr;
   *((int*)(buffer+10)) = push_port;

   count_tcp_traffic(to_user, T35, 12);
   log_write_0x35(to_user, buffer+6);

   g_io_channel_write_chars(to_user->iochannel, buffer,
			    14, NULL, NULL);
   
   g_io_channel_flush(to_user->iochannel, NULL);


   return 1;
}



// TODO: implement
gint
ed2k_send_push_fail(struct UserInfo *user)
{

   return 1;
}



//0x38 in general!
gint 
ed2k_send_string(struct UserInfo *user, char *string, int length)
{
   gchar header[8];


   if (!string)
      return 0;
   
   if (length == -1) 
      length = strlen(string);

   header[0] = 0xe3;
   header[5] = 0x38;

   if (length == -1) 
      length = strlen(string);

   *((int*)           (header+1)) = length + 3;
   *((unsigned short*)(header+6)) = length;
   
   count_tcp_traffic(user, T38, length + 3);

   g_io_channel_write_chars(user->iochannel, header, 8,
			    NULL, NULL);

   g_io_channel_write_chars(user->iochannel, string, length, 
			    NULL, NULL);

   g_io_channel_flush(user->iochannel, NULL);

   
   return 1;
}





// 0x38
gint 
ed2k_send_server_greeting(struct UserInfo *user)
{
   gint  string_size;
   gint  i;
   gchar header[8];
   char **welcome_message;

   
   header[0] = 0xe3;
   header[5] = 0x38;

   welcome_message = get_server_greeting();

   i = 0;
   while (welcome_message[i])
      {
	 string_size = strlen(welcome_message[i]);
	 *((unsigned int*)  (header+1)) = string_size + 3;
	 *((unsigned short*)(header+6)) = string_size;

	 count_tcp_traffic(user, T38, string_size + 3);
	 log_write_0x38(user, string_size);

	 g_io_channel_write_chars(user->iochannel, header, 8,
				  NULL, NULL);
	 g_io_channel_write_chars(user->iochannel, welcome_message[i], 
				  string_size, 
				  NULL, NULL);
	 
	 i++;	 
      }

   g_io_channel_flush(user->iochannel, NULL);	 
	
 
   return 1;
}



// 0x40 id 
gint 
ed2k_send_0x40(struct UserInfo *user)
{
   char packet[10];
   GError *error = NULL;
   int written;

   packet[0] = 0xe3;   
   *((int*)(packet+1)) = 5;
   (char) packet[5] = 0x40;
   *((int*)(packet+6)) = user->id;

   count_tcp_traffic(user, T40, 6);
   log_write_0x40(user);

   g_io_channel_write_chars(user->iochannel, packet, 10, 
			    &written, &error);
   if (!error)
      g_io_channel_flush(user->iochannel, &error);
   
   if (error)      
      printf("[ERROR] error sending 0x40:\n[ ... ] GError: %s\n", 
	     error->message);
   else if (written != 10)
      printf("[ ... ] 0x40 to %s: not all bytes got out\n",
	     user->inet_addr);
   
   return 1;
}






// 0x41
gint
ed2k_send_server_desc(struct UserInfo *user, int keeplock)
{

   char            buffer[1000];
   char           *pos;
   gint            packet_size;
   guint           written;

   guint           name_len, desc_len;
   guchar          *name, *desc;
   

   desc        = get_server_desc();
   name        = get_server_name();
   desc_len    = strlen(desc);
   name_len    = strlen(name);

   buffer[0] = 0xe3;
   buffer[5] = 0x41;

   pos = &(buffer[6]);

   // FUTURE: make some nice server hash

   memset(pos, 0, 16);
   pos += 16;
   
   *(guint*) pos       = rome->pure_ipaddr;
   *(gushort*)(pos+4)  = rome->pure_conn_port;
   pos += 6;

   // no. of tags following
   *(int*) pos = 2;
   pos += 4;

   *(int*) pos      = TAG_NAME_FOLLOWS;
   *(short*)(pos+4) = name_len;
   memcpy(pos+6, name, name_len);
   pos += name_len + 6;

   *(int*) pos      = TAG_SERVER_DESC;
   *(short*)(pos+4) = desc_len;
   memcpy(pos+6, desc, desc_len);
   pos += desc_len + 6;

   
   packet_size = pos - &(buffer[0]);
   *(int*)(buffer+1) = packet_size - 5;

   count_tcp_traffic(user, T41, packet_size - 5);
   log_write_0x41(user);

   if (!keeplock)
      ed2k_data_read_unlock();

   gnet_io_channel_writen(user->iochannel, buffer, packet_size, &written);
   
   g_io_channel_flush(user->iochannel, NULL);


   return 1;
}








// 0x42
gint 
ed2k_send_download_sources(struct UserInfo *user, 
			   struct FileInfo *file,
			   struct SearchResults *results,
			   int keeplock)
{
   char buffer[10000];
   char *pos;
   struct SearchResults *backup;
   int num_results, packet_size;
   

   buffer[0] = 0xe3;
   buffer[5] = 0x42;

   pos = &(buffer[6]);
   
   memcpy(pos, &(file->hash), 16);
   pos += 16;

   // set size later
   pos++;   

   backup = results;
   num_results = 0;
   while (backup)
      {
	 *(int*) pos    = backup->user->id;
	 *(int*)(pos+4) = backup->user->listen_port;
	 pos += 6;
	 backup = backup->next;
	 num_results++;
      }

#ifdef strange_behaviour
   if (num_results == 1)
      {
	 backup = results;

	 *(int*) pos    = backup->user->id;
	 *(int*)(pos+4) = backup->user->listen_port;
	 pos += 6;

	 num_results++;
      }
#endif   

   if (!keeplock)
      ed2k_data_read_unlock();

   packet_size = pos - &(buffer[0]);
   
   *(buffer+22)      = (char)num_results;
   *(int*)(buffer+1) = packet_size - 5;
   
   count_tcp_traffic(user, T42, *((int*)(buffer+1)) );
   log_write_0x42(user, num_results);
   

   g_io_channel_write_chars(user->iochannel, buffer,
			    packet_size, NULL, NULL);

   g_io_channel_flush(user->iochannel, NULL);


   return 1;
}








/* ************************************************************************** *
                                   INIT                                 
   ************************************************************************** */




void ed2k_message_gen_init()
{
   chaos = g_rand_new();  
}


