/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#include <stdio.h>
#include <gnet/gnet.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "structs.h"
#include "serverlist.h"
#include "ed2k_data.h"
#include "ed2k_message_handler.h"
#include "ed2k_message_gen.h"
#include "tcp_connections.h"
#include "udp_handler.h"
#include "timing.h"
#include "maintenance.h"
#include "logging.h"


struct central_struct *rome;

char *romestring = "ROME :-)";

gchar **welcome_message;
gchar **old_welcome_message;







char *
get_server_name()
{
   return "wubbel";
}
 
 
char *
get_server_desc()
{
   return "something to describe";
}




void 
set_default_values()
{
   ///////////////////////////////////////////////////////////////////
   /// logging settings
   ///////////////////////////////////////////////////////////////////

   // this is differential: set only the things *not* set to 
   // zero by default

   // log all five minutes at first
   rome->settings.logging_interval = 5*60*1000;
   rome->settings.new_logging_interval = rome->settings.logging_interval;

   // log user events
   rome->settings.log_user_events = 1;
   
   // log files
   rome->settings.log_files = 1;

   // log searches
   rome->settings.log_text_searches_tcp = 1;
   rome->settings.log_text_searches_udp = 1;


   ///////////////////////////////////////////////////////////////////
   /// network settings
   ///////////////////////////////////////////////////////////////////


}



guint
refresh_server_greeting()
{
   GIOChannel   *greetings_file;
   GError       *error;
   guchar       *buffer;
   guint        *read_bytes;
   guchar      **parsed_buffer;
   gint          i;
   
   
   if (old_welcome_message)
      g_strfreev(old_welcome_message);

   error = NULL;
   greetings_file = g_io_channel_new_file("greeting.txt", "r", &error);
   if (error)
      {
	 printf("[ERROR] error opening welcome file!\n");
	 printf("        keeping old greeting message\n");
	 printf("    --> %s\n", error->message);
	 g_error_free(error);
	 goto __error;
      }
   
   g_io_channel_read_to_end(greetings_file, (gchar**)&buffer, 
			    read_bytes, &error);
   if (error)
      {
	 printf("[ERROR] error reading welcome file!\n");
	 printf("        keeping old greeting message\n");
	 printf("    --> %s\n", error->message);
	 g_error_free(error);
	 g_io_channel_close(greetings_file);
	 goto __error;
      }
   
   (gchar**) parsed_buffer = g_strsplit(buffer, "\n", 1000);
   i = 0;
   while (parsed_buffer[i])
      {
	 g_strstrip(parsed_buffer[i]);
	 i++;
      }
   
   old_welcome_message = welcome_message;
   welcome_message     = (gchar**) parsed_buffer; 
   g_free(buffer);
   
   return 1;

   
 __error:

   return 0;
}



char **
get_server_greeting()
{
   return welcome_message;
}


gint
init_log_files()
{

   if (LOG_ENABLED == 0)
      return 1;

   rome->log_files.trace           = fopen("log_trace.dat", "a");
   rome->log_files.regular_summary = fopen("log_reg_sum.dat", "a");
   rome->log_files.regular_details = fopen("log_reg_det.dat", "a");
   
   if (rome->log_files.trace == NULL
       || rome->log_files.regular_summary == NULL
       || rome->log_files.regular_details ==  NULL)
      {
	 fprintf(stderr, "[FATAL] cannot create log files. aborting\n");
	 exit(-1);
      }

   return 1;
}


gint
read_config_file()
{
   struct UserInfo dummy;
   GIOChannel *configfile;
   GError *error;
   guchar *buffer;
   guchar **commands;
   gsize length;
   int i;

   memset(&dummy, 0, sizeof(struct UserInfo));
   
   dummy.iochannel = g_io_channel_unix_new(1);

   error = NULL;
   configfile = g_io_channel_new_file("cedonkey.cfg", "r", &error);
   if (error)
      {
	 printf("[WARNING] error opening config file, using default values!\n");
	 printf("     -->  error: %s\n", error->message);
	 g_error_free(error);
	 goto __error;
      }
   
   g_io_channel_read_to_end(configfile, (gchar**)&buffer, &length, &error);
   if (error)
      {
	 printf("[WARNING] error reading config file, using default values!\n");
	 printf("     -->  error: %s\n", error->message);
	 g_error_free(error);
	 g_io_channel_close(configfile);
	 goto __error;
      }
   
   commands = (guchar **)g_strsplit(buffer, "\n", 1000);
   g_free(buffer);
   i = 0;
   while (commands[i])
      {
	 g_strstrip(commands[i]);
	 handle_ras_request(&dummy, commands[i]);
	 i++;
      }
   // no. handle_ras_request() FREES the string given (in a threaded context this
   // ... makes perfectly sense). so just free the array, that is enough (more
   // ... is even dangerous! :-)
   // g_strfreev(commands);
   g_free(commands);

   return 1;
   
 __error:
   
   return 0;
}




struct central_struct *
ed2k_init()
{
   rome = g_new0(struct central_struct, 1);

   rome->string = romestring;
   rome->start_time = time(NULL);
   rome->main_loop  = g_main_loop_new(NULL, FALSE);
   
   rome->log_files.unknown_tcp = fopen("./log_unknown_tcp", "a");
   rome->log_files.unknown_udp = fopen("./log_unknown_udp", "a");
   rome->log_files.buggy_pckts = fopen("./log_buggy_pckts", "a");
   
   if (! rome->log_files.unknown_tcp || ! rome->log_files.unknown_udp || 
       ! rome->log_files.buggy_pckts )
      {
	 printf("[CRITICAL] could not open log files!\n");
	 g_assert_not_reached();
      }

   ed2k_message_handler_init();
   ed2k_data_init();
   ed2k_message_gen_init();
   ed2k_serverlist_init();

   welcome_message     = NULL;
   old_welcome_message = NULL;

   set_default_values();

   // from here it HAS to be safe to read & alter config settings

   refresh_server_greeting();
   read_config_file();

   init_log_files();

   recheck_log_files();


   return rome;
}
