/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#include <stdio.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>

#include "littlehelpers.h"
#include "gnet.h"
#include "ed2k_message_handler.h"
#include "ed2k_message_gen.h"
#include "ed2k_data.h"
#include "init.h"
#include "serverlist.h"
#include "timing.h"
#include "logging.h"



/* ************************************************************************** *
                            some globals                            
   ************************************************************************** */


extern struct central_struct *rome;




/* ************************************************************************** *
                         not yet implemented
   ************************************************************************** */


int ras_print_file(struct UserInfo *controller, struct FileInfo *file)
{
   return 0;
}

int ras_print_user(struct UserInfo *sendto, struct UserInfo *user)
{
   return 0;
}

int ras_show_users(struct UserInfo *controller)
{
   return 0;
}

int ras_show_files(struct UserInfo *controller)
{
   return 0;
}


int 
ras_show_files_of_user(struct UserInfo *controller, char *username, char *userid)
{
   return 0;
}
	 


gint
ras_set_ip_of_interface(char *interface)
{
   return 0;
}

gint 
ras_set_ipaddr(char *ipaddr, char *port)
{
   char buffer[16];

   rome->pure_ipaddr    = inet_addr(ipaddr);
   rome->pure_conn_port = atoi(port); 

   // testing purpose ... 
   inet_ntoc(rome->pure_ipaddr, buffer);
   printf("setting ipaddr to %s:%d\n", 
	  buffer, 
	  rome->pure_conn_port);

   return 0;
}




/* ************************************************************************** *
                             some helpers ...
   ************************************************************************** */

gint
truth_eval(char *eval)
{
   if (!eval) 
      return -1;

   if ( (strcmp(eval, "on") == 0)
       || (strcmp(eval, "true") == 0)
       || (strcmp(eval, "1") == 0))
      return 1;
   else if ((strcmp(eval, "off") == 0) 
	    || (strcmp(eval, "false") == 0)
	    || (strcmp(eval, "0") == 0))
      return 0;
   else 
      return -1;
}



gint
match_command(char *command, char *reference, int is, int has_to_be)
{
   if (has_to_be < is)
      return -1;

   if (strcmp(command, reference) == 0)
      return 1;

   return 0;
}
 




/* ************************************************************************** *
                             reply functions
   ************************************************************************** */



void missing_args(struct UserInfo *controller)
{
   char *error = "ERROR: unknown command or missing arguments!\n";
   guint bw;

   gnet_io_channel_writen(controller->iochannel, error, 46, &bw);
}



void ok(struct UserInfo *controller)
{
   char *error = "OK\n";
   guint bw;

   gnet_io_channel_writen(controller->iochannel, error, 3, &bw);
}



void error(struct UserInfo *controller)
{
   char *error = "ERROR IN EXECUTION\n";
   guint bw;

   gnet_io_channel_writen(controller->iochannel, error, 19, &bw);
}



void notyet(struct UserInfo *controller)
{
   char *error = "NOPE: this is not yet implemented, I fear ...\n";
   guint bw;

   gnet_io_channel_writen(controller->iochannel, error, 46, &bw);
}




/* ************************************************************************** *
                                the real fun
   ************************************************************************** */


gint
ras_show_tcpstat(struct UserInfo *controller)
{
   char buffer[1000];
   int written;
   
   // 0x01
   sprintf(buffer, 
	   "INCOMING TCP STATISTICS\n%20s (pkts/bts): %9d / %9d\n\0",
	   "hello 0x01", rome->stat_tcp.got_0x01, 
	   rome->stat_tcp.got_0x01_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x14
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "get srvlist 0x14", rome->stat_tcp.got_0x14, 
	   rome->stat_tcp.got_0x14_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x15
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "offer files 0x15", rome->stat_tcp.got_0x15, 
	   rome->stat_tcp.got_0x15_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x16
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "text search 0x16", rome->stat_tcp.got_0x16, 
	   rome->stat_tcp.got_0x16_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x19
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "query sources 0x19", rome->stat_tcp.got_0x19, 
	   rome->stat_tcp.got_0x19_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x1c
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "push request 0x1c", rome->stat_tcp.got_0x1c, 
	   rome->stat_tcp.got_0x1c_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // unknown
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "unknown", rome->stat_tcp.got_unknown, 
	   rome->stat_tcp.got_unknown_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // alltogether
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "alltogether", rome->stat_tcp.got_messages, 
	   rome->stat_tcp.got_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x32 send server list
   sprintf(buffer, 
	   "OUTGOING TCP STATISTICS\n%20s (pkts/bts): %9d / %9d\n\0",
	   "server list 0x32", rome->stat_tcp.sent_0x32, 
	   rome->stat_tcp.sent_0x32_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x33 search results
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "search results 0x33", rome->stat_tcp.sent_0x33, 
	   rome->stat_tcp.sent_0x33_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x34 server status
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "server status 0x34", rome->stat_tcp.sent_0x34, 
	   rome->stat_tcp.sent_0x34_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x38
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "send string 0x38", rome->stat_tcp.sent_0x38, 
	   rome->stat_tcp.sent_0x38_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x40 send id
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "send id 0x40", rome->stat_tcp.sent_0x40, 
	   rome->stat_tcp.sent_0x40_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x41 send id
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "server info 0x41", rome->stat_tcp.sent_0x41, 
	   rome->stat_tcp.sent_0x41_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x42 send id
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "send dl sources 0x42", rome->stat_tcp.sent_0x42, 
	   rome->stat_tcp.sent_0x42_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // alltogether
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "alltogether", rome->stat_tcp.sent_messages, 
	   rome->stat_tcp.sent_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);


   g_io_channel_flush(controller->iochannel, NULL);
			  
   

   return 1;
}





gint
ras_show_udpstat(struct UserInfo *controller)
{
   char buffer[1000];
   int written;
   
   // 0x96 ping
   sprintf(buffer, 
	   "INCOMING UDP STATISTICS\n%20s (pkts/bts): %9d / %9d\n\0",
	   "ping 0x96", rome->stat_udp.got_0x96, 
	   rome->stat_udp.got_0x96_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x98 search
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "text search 0x98", rome->stat_udp.got_0x98, 
	   rome->stat_udp.got_0x98_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x9a query sources
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "get sources 0x9a", rome->stat_udp.got_0x9a, 
	   rome->stat_udp.got_0x9a_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x9c push request
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "push request 0x9c", rome->stat_udp.got_0x9c, 
	   rome->stat_udp.got_0x9c_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0xa0 req serverlist
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "req servlist 0xa0", rome->stat_udp.got_0xa0, 
	   rome->stat_udp.got_0xa0_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // unknown
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "unknown", rome->stat_udp.got_unknown, 
	   rome->stat_udp.got_unknown_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // alltogether
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "alltogether", rome->stat_udp.got_packets, 
	   rome->stat_udp.got_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x97 pong
   sprintf(buffer, 
	   "OUTGOING UDP STATISTICS\n%20s (pkts/bts): %9d / %9d\n\0",
	   "pong 0x97", rome->stat_udp.sent_0x97, 
	   rome->stat_udp.sent_0x97_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x99 search result
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "search result 0x99", rome->stat_udp.sent_0x99, 
	   rome->stat_udp.sent_0x99_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0x9b return download sources
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "send dl sources 0x9b", rome->stat_udp.sent_0x9b, 
	   rome->stat_udp.sent_0x9b_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // 0xa1 return download sources
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "server list 0xa1", rome->stat_udp.sent_0xa1, 
	   rome->stat_udp.sent_0xa1_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);

   // alltogether
   sprintf(buffer, "%20s (pkts/bts): %9d / %9d\n\0",
	   "alltogether", rome->stat_udp.sent_packets, 
	   rome->stat_udp.sent_bytes);

   gnet_io_channel_writen(controller->iochannel, buffer, 
			  strlen(buffer), &written);


   g_io_channel_flush(controller->iochannel, NULL);
			  

   return 1;
}




int 
ras_add_server_file(char *file)
{
   GIOChannel  *server_file;
   char        *buffer;
   char       **parsed_buffer;
   char       **parsed_string;
   gint         read_bytes;
   GError      *error;
   gint         i;

   error = NULL;
   server_file = g_io_channel_new_file(file, "r", &error);
   if (error)
      {
	 printf("[ERROR] add server file: %s\n", error->message);
	 g_error_free(error);
	 goto __error;
      }
   
   g_io_channel_read_to_end(server_file, &buffer, &read_bytes, &error);
   if (error)
      {
	 printf("[ERROR] add server file: %s\n", error->message);
	 g_error_free(error);
	 g_io_channel_close(server_file);
	 goto __error;
      }

   g_io_channel_close(server_file);
   
   parsed_buffer = g_strsplit(buffer, "\n", 2000);
   i = 0;
   while (parsed_buffer[i])
      {
	 g_strstrip(parsed_buffer[i]);
	 parsed_string = g_strsplit(parsed_buffer[i], " ", 10);
	 if (parsed_string[0])
	    server_add_cc(parsed_string[1], parsed_string[2]);
	 g_strfreev(parsed_string);
	 i++;
      }
   
   g_strfreev(parsed_buffer);
   
   
   return 1;

 __error: 
   return 0;
}



int
ras_show_info(struct UserInfo *controller)
{
   char buffer[500];
   char ipaddress[16];
   guint written;

   sprintf(buffer,
	   "%25s: %6d\n%25s: %6d / %6d (tracked/counted)\n%25s: %6d / %6d (tracked/counted)\n%25s: %6d\n\0",
	   "num_files_all", ed2k_data_get_all_files_count(),
	   "num_files_different", ed2k_data_get_different_files_count(),
	   ed2k_data_count_different_files(),
	   "num_users", ed2k_data_get_users_count(),
	   ed2k_data_count_users(),
	   "num_users_lowid", ed2k_data_get_lowid_count());
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);

   sprintf(buffer, "%25s: %6d (counted)\n\n\0",
	   "num_servers", server_get_count());
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);
   
   sprintf(buffer, "%25s: %6d\n\0", "debug: user not found", 
	   rome->debugdata.user_not_found);
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);

   sprintf(buffer, "%25s: %6d\n\n\0", "debug: socket error", 
	   rome->debugdata.socket_error);
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);

   sprintf(buffer, "%25s: %6d\n\0", "debug: verbose", 
	   rome->settings.verbose);
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);

   sprintf(buffer, "%25s: %6d\n\n\0", "debug: debuglevel", 
	   rome->settings.debuglevel);
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);

   inet_ntoc(rome->pure_ipaddr, ipaddress);
   sprintf(buffer, "%25s: %s (%u)\n%25s: %d\n\n\0", "using ipaddress", 
	   ipaddress, rome->pure_ipaddr,
	   "public server", rome->settings.public);
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);

   sprintf(buffer, "%25s: %u min\n\n\0", "uptime", 
	   (unsigned int)(time(NULL) - rome->start_time) / 60);
   gnet_io_channel_writen(controller->iochannel, buffer, strlen(buffer), &written);

   g_io_channel_flush(controller->iochannel, NULL);


   return 1;
}





int
ras_set_tcpworkers(char *workers)
{
   rome->tcp_workers = atoi(workers);
   return 1;
}


int
ras_set_udpworkers(char *workers)
{
   rome->udp_workers = atoi(workers);
   return 1;
}






/* ************************************************************************** *
                     the thread in which this is running
   ************************************************************************** */



int handle_ras_request(struct UserInfo *controller, char *raw_request)
{
   int count;
   char                     **request;
   int value;

   g_strstrip(raw_request);
   
   if (! *raw_request)
      return 1;
   else if (*raw_request == '#')
      return 1;
   
   request = g_strsplit(raw_request, " ", 4);
   g_free(raw_request);
	 
   // should work also with a glib parsed thing ... :-)
   count = count_parsed_strings(request);
   


   if ( count && strcmp(request[0], "show") == 0)
      {
	       
	 if (count < 2) missing_args(controller);
	       
	 else if (strcmp(request[1], "users") == 0)
	    {
	       if (ras_show_users(controller))
		  ok(controller);
	       else 
		  error(controller);
	    }
	 else if (strcmp(request[1], "servers") == 0)
	    {
	       notyet(controller);
	    }
	 else if (strcmp(request[1], "files") == 0) 
	    {
	       notyet(controller);
/* 	       if (count == 2) */
/* 		  ras_show_files(controller); */
/* 	       else if (count == 4) { */
/* 		  if (strcmp(request[2], "user") == 0)  */
/* 		     ras_show_files_of_user(controller, request[3], NULL); */
/* 		  if (strcmp(request[2], "uid") == 0) */
/* 		     ras_show_files_of_user(controller, NULL, request[3]); */
	    }
	 else if (strcmp(request[1], "info") == 0)
	    {
	       if (ras_show_info(controller))
		  ok(controller);
	       else
		  error(controller);
	    }
	 else if (strcmp(request[1], "tcpstat") == 0)
	    {
	       if (ras_show_tcpstat(controller))
		  ok(controller);
	       else
		  error(controller);
	    }
	 else if (strcmp(request[1], "udpstat") == 0)
	    {
	       if (ras_show_udpstat(controller))
		  ok(controller);
	       else
		  error(controller);
	    }
	 else if (strcmp(request[1], "nothing") == 0)
	    {
	       ok(controller); // :-)
	    }
      }
   else if ( count && strcmp(request[0], "set") == 0)
      {
	 if (count < 2) missing_args(controller);	       

	 else if (strcmp(request[1], "ipaddr") == 0)
	    {
	       if (count < 4) missing_args(controller);
	       else {
		  if (strcmp(request[2], "if") == 0)
		     ras_set_ip_of_interface(request[4]);
		  else
		     ras_set_ipaddr(request[2], request[3]);
		  make_server_public(NULL);
	       }
	    }
	 else if (strcmp(request[1], "tcpworkers") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else 
		  if (ras_set_tcpworkers(request[2])) ok(controller);
		  else  error(controller);
	    }
	 else if (strcmp(request[1], "udpworkers") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else 
		  if (ras_set_udpworkers(request[2]));
		  else  error(controller);
	    }
	 else if (strcmp(request[1], "welcome_message") == 0)
	    {
	       if (refresh_server_greeting())
		  ok(controller);
	       else
		  error(controller);
	    }
	 else if (strcmp(request[1], "public") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.public = value;
		     ok(controller);
		  }
	       }
	    }
	 else if (strcmp(request[1], "debuglevel") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  rome->settings.debuglevel = atoi(request[2]);
		  ok(controller);
	       }
	    }

	 else if (strcmp(request[1], "verbose") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  rome->settings.verbose = atoi(request[2]);
		  ok(controller);
	       }
	    }


	 
	 //////////////////////////////////////////////////////
	 /// logging settings
	 //////////////////////////////////////////////////////



	 else if (strcmp(request[1], "log_everything") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_everything = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_user_events") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_user_events = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_files") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_files = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_text_searches_udp") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_text_searches_udp = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_text_searches_tcp") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_text_searches_tcp = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_recv_pings") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_recv_pings = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_recv_pongs") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_recv_pongs = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_query_sources_udp") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_query_sources_udp = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  
	 else if (strcmp(request[1], "log_query_sources_tcp") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  value = truth_eval(request[2]);
		  if (value == -1) error(controller);
		  else {
		     rome->settings.log_query_sources_tcp = value;
		     recheck_log_files();
		     ok(controller);
		  }
	       }
	    }  

      }
   else if ( count && strcmp(request[0], "add") == 0)
      {
	 if (count < 2) missing_args(controller);	       

	 else if (strcmp(request[1], "server") == 0)
	    {
	       if (count < 4) missing_args(controller);
	       else
		  if (server_add_cc(request[2], request[3]))
		     ok(controller);
		  else
		     error(controller);
	    }

	 if (strcmp(request[1], "serverfile") == 0)
	    {
	       if (count < 3) missing_args(controller);
	       else {
		  if (ras_add_server_file(request[2])) {
		     ok(controller);
		     make_server_public(NULL);
		  }
		  else
		     error(controller);
	       }
	    }
      }
   else if ( count && strcmp(request[0], "debug") == 0)
      {
	 if (count < 2) missing_args(controller);
	 
	 else if (strcmp(request[1], "dump") == 0)
	    {
	       debug_dump();
	    }
      }


   return 1;
}





gint 
handle_ras_request_thread(GAsyncQueue *queue)
{
   // struct central_struct *_rome;
   struct UserInfo *controller;
   char            *raw_request;

   // _rome = g_async_queue_pop(queue);
   // rome  = _rome;
   
   printf("handle_ras_request_thread: ready. (rome = %s)\n", rome->string);
   
   while (1)
      {
	 controller  = g_async_queue_pop(queue);
	 raw_request = g_async_queue_pop(queue);
	 
	 handle_ras_request(controller, raw_request);
      } 
   
   g_assert_not_reached();
}
