/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */


/* ************************************************************************** *
   WARNING: 
   locking here has to be used to guarantee consistent data trees.
   but LOCKING HAS TO BE USED FROM THE OUTSIDE!!
   every call to server_get_all() has to be bedded in server_lock()
   and server_unlock() calls
 * ************************************************************************** */

#include <stdlib.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <glib-2.0/glib.h>

#include "structs.h"
#include "glib-2.0/glib.h"
#include "littlehelpers.h"




/* ************************************************************************** * 
   wow. globals ...                     
 * ************************************************************************** */



GTree                    *ServerTree;

gint                     *list_changed;

struct ServerList        *first_server;

struct ServerList        *next_server;

GMutex                   *mutex;




/* ************************************************************************** * 
   internal functions only
 * ************************************************************************** */

gint
server_add_internal(struct ServerList *new_server)
{
   g_tree_insert(ServerTree, &(new_server->IP), new_server);

   (*list_changed)++;
   
   return 1;
}



gint
compare_servers(guint *server_ip_1, guint *server_ip_2)
{
   if (*server_ip_1 > *server_ip_2)
      return 1;
   else if (*server_ip_1 < *server_ip_2)
      return -1;
   else 
      return 0;
}



gboolean
server_chain_servers(guint *ip1, struct ServerList *server, 
		     struct ServerList **serverlist)
{
   // this is quite clever, I think ;-)) 

   // is this the first one?
   if (! *serverlist)
      {
	 // the first (and only??) server
	 first_server = server;
	 server->next = server;
	 server->prev = server;
      }
   else
      {
	 (*serverlist)->next = server;
	 server->prev = *serverlist;
	 
	 // to create ringlist
	 server->next = first_server;
	 first_server->prev = server;
      }

   *serverlist = server;


   return FALSE;
}





struct ServerList *
server_present_internal(guint ipaddr)
{
   guint *orig_ip;
   gpointer data;
   
   if (g_tree_lookup_extended(ServerTree, (gpointer) &ipaddr, 
			      (gpointer) &orig_ip, &data))
      return data;
   else
      return NULL;
}


guint
server_c_to_n(char *servername)
{
   return inet_addr(servername);
}


void
server_n_to_c(guint ip, char *address)
{
   guchar *val;

   val = (guchar*) &ip;

   sprintf(address, "%d.%d.%d.%d\0", 
	   val[0],
	   val[1],
	   val[2],
	   val[3]);
}



/* ************************************************************************** * 
   some statistic functions
 * ************************************************************************** */

gint
server_get_count()
{
   return g_tree_nnodes(ServerTree);
}







/* ************************************************************************** */
/*                           locking (also public)                            */
/* ************************************************************************** */

void
server_lock()
{
   g_mutex_lock(mutex);
}

void
server_unlock()
{
   g_mutex_unlock(mutex);
}



/* ************************************************************************** * 
   public functions
 * ************************************************************************** */



gint
ed2k_serverlist_init()
{
   ServerTree = g_tree_new((GCompareFunc) compare_servers);

   first_server = NULL;
   next_server  = NULL;

   list_changed = g_new0(gint, 1);
   
   mutex = g_mutex_new();

   
   return 1;
}




struct ServerList *
server_get_all()
{
   struct ServerList *serverlist;
   struct ServerList *temp;
   int                i;
   
   serverlist = NULL;

   if (list_changed || (!first_server) )
      {
	 g_tree_foreach(ServerTree, 
			(GTraverseFunc)server_chain_servers, 
			&serverlist);
      }
	 
   *list_changed = 0;
   
   temp = next_server;

   if (!temp) 
      {
	 temp = first_server;
	 next_server = first_server;
      }

   if (!temp)
      {
	 return NULL;
      }

   // just do down the serverlist! nothning more!
   serverlist = next_server;
   for (i=0; i < MAX_SERVERS_TO_SEND; i++)
      {
	 next_server = next_server->next;
	 if (next_server == serverlist)
	    i = MAX_SERVERS_TO_SEND;
      }


   if (temp)
      return temp;
   else
      return first_server;
}



gint
server_add_cc(guchar *ip, guchar *port)
{
   struct ServerList      *new_server;
   gushort                 num_port;
   guint                   num_ip;

   
   if (server_present_internal( *((guint*)(ip))  ) ) 
      return 0;

   new_server = g_new0(struct ServerList, 1);
   num_port               = atoi(port);
   num_ip                 = inet_addr(ip);
   new_server->IP         = num_ip;
   new_server->port       = num_port;
   new_server->first_seen = time(NULL);

   new_server->addr = gnet_inetaddr_new(ip, num_port);

   server_add_internal(new_server);


   return 1;
}




gint
server_add_nn(guint ip, gint port)
{
   struct ServerList       *new_server;
   char                     ipbuf[16];
   
   if (server_present_internal(ip))
      return 0;

   new_server = g_new0(struct ServerList, 1);
   server_n_to_c(ip, ipbuf);
   new_server->IP         = ip;
   new_server->port       = port;
   new_server->first_seen = time(NULL);

   new_server->addr = gnet_inetaddr_new(ipbuf, port);
   
   server_add_internal(new_server);
   

   return 1;
}





gint
server_delete(guint ipaddr)
{
   struct ServerList *orig_server;

   orig_server = server_present_internal(ipaddr);
   
   if (!orig_server)
      {
	 fprintf(stdout, "[WUPS] going to delete unknown server. how?\n");
	 return 0;
      }


   if (orig_server->next == orig_server)
      // the only server in the list, seldom but possible
      {
	 first_server = NULL;
	 next_server  = NULL;
      }
   else
      {
	 if (orig_server == first_server)
	    {
	       first_server = first_server->next;
	    }
	 else if (orig_server == next_server)
	    {
	       next_server = next_server->next;
	    }
      }
   
   // dont forget to really remove (!) the thing

   orig_server->prev->next = orig_server->next;
   orig_server->next->prev = orig_server->prev;

   g_tree_remove(ServerTree, &(orig_server->IP));

   gnet_inetaddr_delete(orig_server->addr);
   g_free(orig_server);

   
   return 1;
}



gint
server_present(guint *ipaddr)
{
   if (server_present_internal(*ipaddr))
      return 1;
   else
      return 0;
}



gint
server_register_ping(GInetAddr* addr, guint challenge)
{
   struct ServerList *server;
   guint              ipaddr;

   ipaddr = inet_addr(gnet_inetaddr_get_canonical_name(addr));
   
   server = server_present_internal(ipaddr);

   if (server)
      {
	 if (server->challenge)
	    server->num_timeouts++;
	 
	 server->challenge  = challenge;
	 server->ping_start = time(NULL);

	 // so this one is taken out of the chain if list requested
	 (*list_changed)++;

	 return 1;
      }
   else
      {
	 printf("[ERROR] how can I ping an unknown server??\n");
	 return 0;
      }
}



gint
server_register_ping_struct(struct ServerList *server, guint challenge)
{
   if (server->challenge)
      server->num_timeouts++;
   
   server->challenge  = challenge;
   server->ping_start = time(NULL);

   (*list_changed)++;

   
   return 1;
}



gint
server_deregister_ping(GInetAddr *address, guint *values)
{
   struct ServerList *server;
   guint              ipaddr;
   
   ipaddr = inet_addr(gnet_inetaddr_get_canonical_name(address));
   
   server = server_present_internal(ipaddr);

   if (server)
      {

	 if (server->challenge != *values)
	    {
	       printf("[WUPS] got PONG from server with wrong challenge!\n");
	       printf("       deleting it.\n");
	       server_delete(ipaddr);
	       return 0;
	    }

	 server->last_ping_duration = time(NULL) - server->ping_start; 
	 server->challenge          = 0;
	 server->ping_start         = 0;
	 
	 server->num_users          = values[1];
	 server->num_files          = values[2];

	 (*list_changed)++;
	 
	 return 1;
      }
   else
      {
	 printf("[WUPS] got PONG from unknown server ...\n");
	 return 0;
      }

}


gint
server_check_pings()
{


   return 1;
}
