/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#ifndef _structs
#define _structs 1

#include <gnet/gnet.h>
#include <glib.h>
#include <time.h>
#include <stdio.h>



/* ************************************************************************** *
                              other definitions
   ************************************************************************** */

#define LOG_ENABLED 1

#define UDP_IN  1
#define UDP_OUT 2



/* ************************************************************************** *
                              tag definitions                                
   ************************************************************************** */


// note: 2,1,0,1 (16777474) means "name", NOT "FILEname" or "USERname". 
// so its contextdependant!

#define TAG_NAME_FOLLOWS          16777474              // 02 01 00 01, string
#define TAG_FILE_NAME             TAG_NAME_FOLLOWS      //   s.o.
#define TAG_USERNAME              TAG_NAME_FOLLOWS      //   s.o.
#define TAG_SERVER_DESC           184549634             // 02 01 00 0b, string
#define TAG_CLIENT_PORT           251658499             // 03 01 00 0f, int
#define TAG_CLIENT_VERSION        285212931             // 03 01 00 11, int
#define TAG_FILE_SIZE             33554691              // 03 01 00 02, int
#define TAG_FILE_AVAIL            352321795             // 03 01 00 15, int


#define ED2K_CONNECTION_PORT      4661
#define RAS_CONNECTION_PORT       5000



/* ************************************************************************** *
                           packet type definitions                          
   ************************************************************************** */


#define T01 0x01 // hello
#define T14 0x14 // req. srv.list
#define T15 0x15 // publish files
#define T16 0x16 // text search
#define T19 0x19 // source search
#define T1c 0x1c // push req.


#define T32 0x32 // srv.list
#define T33 0x33 // txt srch. results
#define T34 0x34 // srv. status
#define T35 0x35 // push command
#define T38 0x38 // write string
#define T40 0x40 // id transmit
#define T41 0x41 // server info
#define T42 0x42 // download sources

#define U96 0x96 // ping
#define U97 0x97 // pong
#define U98 0x98 // text search
#define U99 0x99 // search result
#define U9a 0x9a // search sources
#define U9b 0x9b // download sources
#define U9c 0x9c // push request
#define U9e 0x9e // push failed
#define Ua0 0xa0 // server announce
#define Ua1 0xa1 // serverlist
#define Ua2 0xa2 // req. srv. info
#define Ua3 0xa3 // server info
#define Ua4 0xa4 // req. serverlist

#define T00 0xF0 // tcp unknown 
#define U00 0xF1 // udp unknown 





/* ************************************************************************** *
                              other definitions                              
   ************************************************************************** */

#define MAX_SERVER_ANNOUNCE   50
#define MAX_SERVERS_TO_SEND   50
#define MAX_RESULTS_TO_SEND  254
#define MAX_SERVERS_TO_PING   25



/* ************************************************************************** *
                         statistical & logging structs                  
   ************************************************************************** */

struct udp_stat {
   int got_bytes;
   int got_packets;

   int sent_bytes;
   int sent_packets;

   int got_0x96;
   int got_0x96_bytes;
   int got_0x97;
   int got_0x97_bytes;
   int got_0x98;
   int got_0x98_bytes;
   int got_0x9a;
   int got_0x9a_bytes;
   int got_0x9c;
   int got_0x9c_bytes;
   int got_0xa0;
   int got_0xa0_bytes;
   int got_0xa1;
   int got_0xa1_bytes;
   int got_0xa2;
   int got_0xa2_bytes;
   int got_0xa3;
   int got_0xa3_bytes;
   int got_0xa4;
   int got_0xa4_bytes;

   int got_unknown;
   int got_unknown_bytes;

   int sent_0x96;
   int sent_0x96_bytes;
   int sent_0x97;
   int sent_0x97_bytes;
   int sent_0x99;
   int sent_0x99_bytes;
   int sent_0x9b;
   int sent_0x9b_bytes;
   int sent_0x9e;
   int sent_0x9e_bytes;
   int sent_0xa0;
   int sent_0xa0_bytes;
   int sent_0xa1;
   int sent_0xa1_bytes;
   int sent_0xa2;
   int sent_0xa2_bytes;
   int sent_0xa3;
   int sent_0xa3_bytes;
   int sent_0xa4;
   int sent_0xa4_bytes;

   int hash_search_successful;
   int hash_search_results;

   int text_search_successful;
   int text_search_results;

   void *rw_lock;
};



struct tcp_stat {
   int con_reqs;
   int con_drops;

   int got_messages;
   int got_bytes;

   int sent_messages;
   int sent_bytes;

   int got_0x01;
   int got_0x01_bytes;
   int got_0x14;
   int got_0x14_bytes;
   int got_0x15;
   int got_0x15_bytes;
   int got_0x16;
   int got_0x16_bytes;
   int got_0x19;
   int got_0x19_bytes;
   int got_0x1c;
   int got_0x1c_bytes;

   int got_unknown;
   int got_unknown_bytes;
   
   int sent_0x32;
   int sent_0x32_bytes;
   int sent_0x33;
   int sent_0x33_bytes;
   int sent_0x34;
   int sent_0x34_bytes;
   int sent_0x35;
   int sent_0x35_bytes;
   int sent_0x36;
   int sent_0x36_bytes;
   int sent_0x38;
   int sent_0x38_bytes;
   int sent_0x40;
   int sent_0x40_bytes;
   int sent_0x41;
   int sent_0x41_bytes;
   int sent_0x42;
   int sent_0x42_bytes;

   int hash_search_successful;
   int hash_search_results;

   int text_search_successful;
   int text_search_results;

   void* rw_lock;
};
   

struct user_stat
{
   int bogus;
};


struct log_files_struct
{
   FILE *unknown_tcp;      // log of unknown packets
   FILE *unknown_udp;      // log of unknown packets
   FILE *buggy_pckts;      // log of buggy packets
   FILE *user_events;      // log for all user events
   FILE *files;            // file drops out of net
   FILE *searches;         // log all search strings!
   // FILE *alltogether;   // ??

   FILE *trace;            // for everything log ... :-)

   FILE *regular_summary;  // most important: regular summary 
   FILE *regular_details;  // most important: regular details
};



struct debug_struct
{
   gint user_not_found;
   gint socket_error;
   gint unfriendly_hangup;
   gint dontknowwhat;
   gint doubledelete;
};



struct settings_struct
{
   gint debuglevel;
   gint verbose;
   gint public;

   // logging settings
   
   gint logging_interval;
   gint new_logging_interval;

   gint log_user_events;
   gint log_everything;
   gint log_files;

   // ping and pong
   gint log_recv_pings;
   gint log_sent_pings;
   gint log_recv_pongs;
   gint log_sent_pongs;
   
   // server description
   gint log_recv_serverdescs_tcp;
   gint log_sent_serverdescs_tcp;
   gint log_recv_serverdescs_udp;
   gint log_sent_serverdescs_udp;

   // query sources
   gint log_query_sources_answers_udp;
   gint log_query_sources_answers_tcp;
   gint log_query_sources_udp;
   gint log_query_sources_tcp;
   
   // text searches
   gint log_text_searches_udp;
   gint log_text_searches_tcp;
   gint log_text_search_answers_udp;
   gint log_text_search_answers_tcp;
   
   // announces of servers
   gint log_sent_announces;
   gint log_recv_announces;

   // serverlist requests
   gint log_recv_srvlist_requests_udp;
   gint log_sent_srvlist_requests_udp;
   gint log_recv_srvlist_requests_tcp;
   gint log_sent_srvlist_requests_tcp;

   // serverlists 
   gint log_sent_serverlists_udp;
   gint log_recv_serverlists_udp;
   gint log_sent_serverlists_tcp;
   gint log_recv_serverlists_tcp;

   // server stati
   gint log_sent_serverstatus_tcp;
   gint log_sent_serverstatus_udp;
   gint log_recv_serverstatus_udp;

   // push requests & commands
   gint log_pushrequests_tcp;
   gint log_pushrequests_udp;
   gint log_pushcommands_tcp;
   gint log_pushcommands_udp;

   // strings, ids & id changes
   gint log_strings;
   gint log_ids;
   gint log_id_changes;
};




/* ************************************************************************** *
                              data core structs                              
   ************************************************************************** */


struct DescInfo
{
   char  *string;
   int    strlen; 
   int    ref_count;
   int   *searched;
   GTree *files;
};


struct FileInfo
{
   // descriptions are in users' info!
   GTree               *owners; 
   gchar                hash[16];
   guint                size;
   guint                count;
   guint                type;
   char                *add_info;
   guint                info_size;
   
};



struct UserInfo
{
   GTcpSocket          *socket;

   GIOChannel*          iochannel;
   guint                watch_flags;
   GSource             *gsource;
   struct PollContext  *context;
   guint                watch;
   guint                n;

   guint                ipaddr;
   guint                port;
   gchar               *inet_addr;
   
   guint                id;
   guint                listen_port;

   gchar                hash[16];
   gchar                name[30];
   guint                name_length;
   time_t               start_time;
   guint                version;
   guint                shared_files_count;
   guint                shared_files_size;
   GTree               *shared_files;
   // 10=state checking, 1=open, 2=firewalled
   int                  state;
   
   gint                 msg_count;
   gint                 msg_size; 
   struct tcp_stat      stat_tcp;
};



struct ServerList {
   
   GInetAddr *addr;

   // again: dont separate following two values!
   guint IP;
   gushort port;

   time_t first_seen;

   gint   num_users;
   gint   num_files;
   
   char  *name;
   char  *desc;

   time_t ping_start;
   guint  challenge;
   gint   num_timeouts;
   guint  last_ping_duration;

   gint   sent_announce;

   struct ServerList *next;
   struct ServerList *prev;
};



struct PollContext
{
   GThread      *thread;
   GMainContext *context;
   GMainLoop    *loop;

   gint          count;

   struct PollContext *next;
   struct PollContext *prev;
};




/* ************************************************************************** *
                         structures protocol handling                   
   ************************************************************************** */

// tagREFstruct, cause it JUST POINTS on OTHER BUFFERS LOCATIONS!!!
// it does not hold any tags by itself!

struct TagRefStruct {
   guchar    type;
   gushort   T1_len;
   guchar   *T1;
   gushort   T2_len;
   guchar   *T2;
};




/* ************************************************************************** *
                         structures for data passing                    
   ************************************************************************** */



struct SearchResults {
   struct FileInfo       *file;    
   struct DescInfo       *desc;    
   struct UserInfo       *user;
   struct SearchResults  *next;
};

struct SearchBaseStruct {
   GTree                 *strees[4];
   int                    already_found;
   struct DescInfo       *current_desc;
   struct SearchResults  *results;
};

struct message_handle_struct
{
   struct UserInfo    *client_info;
   gchar              *buffer;
   guint               size;
   GInetAddr          *udp_address;
   GUdpPacket         *udp_packet;
};



/* ************************************************************************** *
                    central struct - the one which holds it all    
   ************************************************************************** */




struct central_struct {
   // holds a smiley, this is important :-))
   char                    *string;
   
   // when we began
   time_t                   start_time;

   // the tcp workers management
   gint                     tcp_workers;
   struct PollContext      *contexts;
   GMainLoop               *main_loop;

   // thread pool for udp work
   gint                     udp_workers;
   GThread                 *udp_thread;
   GThreadPool             *udp_pool;

   // the udp socket to send packets to ... :-)
   GUdpSocket              *udp_socket;
   
   // server network stuff
   // these two should not be needed. :-)
   GInetAddr               *running_addr;
   GInetAddr               *outside_addr;
   // DONT SEPARATE THE FOLLOWING TWO!!!
   guint                    pure_ipaddr;
   guint                    pure_conn_port;
   guint                    pure_ras_port;
   gchar                   *hostname;
   
   // server behaviour stuff
   gint                     max_connections;
   GInetAddr               *seed_addr;

   // the statistic stuff ... 
   struct udp_stat          stat_udp;
   struct tcp_stat          stat_tcp;
   struct log_files_struct  log_files;
   struct settings_struct   settings;

   struct debug_struct      debugdata;

   struct user_stat         stat_user;
};




#endif
