/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#include <stdio.h>
#include <gnet.h>
#include <arpa/inet.h>
#include <stdlib.h>

#include "ed2k_data.h"
#include "structs.h"
#include "structs.h"
#include "maintenance.h"
#include "littlehelpers.h"
#include "tcp_listeners.h"
#include "settings.h"




struct TPStruct {
   gpointer one;
   gpointer two;
   gpointer three;
};





/* the whole thing is three threads: 
 *   - one for waiting on status changes
 *   - one for initializing connections
 *   - one for reading the tcp messages and passing them to the 
 *     processing functions
 */






/* ************************************************************************** *
                             some globals                               
   ************************************************************************** */


extern struct central_struct *rome;

guint   server_address;
gushort server_port;

struct PollContext *current_context;



/* ************************************************************************** *
                          some helper functions                         
   ************************************************************************** */




gushort 
tcp_get_server_port()
{
   return rome->pure_conn_port;
}


gushort 
tcp_get_server_udpport()
{
   return rome->pure_conn_port + 4;
}







/* ************************************************************************** *
                                    start                                   
   ************************************************************************** */





gboolean get_ras_command(GIOChannel *iochannel, GIOCondition condition, 
			 gpointer data)
{
   
   // always needed
   struct UserInfo *client_info;

   // for the core functionality
   GError    *error = NULL;
   guint      read_bytes;
   gchar     *buffer;


   client_info = (struct UserInfo*) data;


   if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
   {
	 fprintf(stdout, "[ERROR] socket: RAS client socket error: (%s:%d). closing socket\n", 
		 client_info->inet_addr, client_info->port);
	 goto __close_socket;
   }
   
   if (! (condition & G_IO_IN)) goto __next;

   g_io_channel_read_line(iochannel, &buffer, &read_bytes, 
				     NULL, &error);
   
   if (read_bytes == 0)
      // channel closed
      {
	 goto __close_socket;
      }

   if (error) 
   {
	 printf("[ERROR] error occurred when reading ed2k message - closing socket\n");
	 printf("[ERROR] GError: %s\n", error->message);
	 g_clear_error(&error);
	 goto __close_socket;
   }

   handle_ras_request(client_info, buffer);

 __next:

   return TRUE;


 __close_socket:
   remove_connection(client_info);
   
   return FALSE;


}



gboolean tcp_ras_connection(GTcpSocket *server, GTcpSocket *client, gpointer data)
{
   GError          *error;
   guint            port, flags;
   GInetAddr       *addr;
   gchar           *name;
   GIOChannel      *client_io;
   struct UserInfo          *client_info;

   // generate some debug output
   addr = gnet_tcp_socket_get_inetaddr(client);
   name = gnet_inetaddr_get_canonical_name(addr);
   port = gnet_inetaddr_get_port(addr);
   
#ifdef serious_debug      
   printf("[DEBUG] RAS connection from: %s:%d\n", name, port);
#endif
   
   client_io               = gnet_tcp_socket_get_iochannel(client);
   
   client_info             = g_new0(struct UserInfo, 1);
   client_info->socket     = client;
   client_info->start_time = time(NULL);
   client_info->iochannel  = client_io;
   client_info->ipaddr     = inet_addr(name);
   client_info->port       = port;
   client_info->inet_addr  = name;
   
   
   flags = G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL;

   error = NULL;
   // gnet & glib2 inconsistencies :-)
   if (! g_io_channel_get_buffered(client_io) )
      g_io_channel_set_buffered(client_io, TRUE);
   else 
      g_io_channel_set_encoding(client_io, NULL, &error);
   if (error)
      printf("[ERROR] [01] GError: %s\n", error->message);
   
   client_info->watch = g_io_add_watch(client_io, flags, 
				       get_ras_command, client_info);
   
   return TRUE;
}









/* *************************************************************************** *
 * edonkey tcp stuff - connections                                             *
 * *************************************************************************** */

gboolean 
tcp_edonkey_connection(GTcpSocket *server, GTcpSocket *client, gpointer data)
{
   GSource          *source;
   GError           *error;
   guint             port, flags;
   GInetAddr        *addr;
   gchar            *name;
   GIOChannel       *client_io;
   struct UserInfo  *client_info;

   
   // log the connection request
   rome->stat_tcp.con_reqs++;


   // generate some debug output
   addr = gnet_tcp_socket_get_inetaddr(client);
   name = gnet_inetaddr_get_canonical_name(addr);
   port = gnet_inetaddr_get_port(addr);
   
#ifdef serious      
   if (cvl() > 0)
      printf("[connect ] ? %s:%d context %d\n", 
	     name, port, current_context->count);
#endif
   
   // this is a server socket. refer to gnet docs ("if you can read from it,
   // ... there's a connection waiting")
   client_io = gnet_tcp_socket_get_iochannel(client);
   
   /* client info is a struct holding eDonkey network information *
    * and simple internet information of the certain client       */
   
   client_info             = g_new0(struct UserInfo, 1);
   client_info->socket     = client;
   client_info->start_time = time(NULL);
   client_info->iochannel  = client_io;
   client_info->ipaddr     = inet_addr(name);
   client_info->port       = port;
   client_info->inet_addr  = name;
   
   
   error = NULL;
   // gnet & glib2 inconsistencies :-)
   if (! g_io_channel_get_buffered(client_io) )
      g_io_channel_set_buffered(client_io, TRUE);
   else 
      g_io_channel_set_encoding(client_io, NULL, &error);
   if (error)
      printf("[ ERROR! ] [01] GError: %s\n", error->message);
   
   flags = G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL;
   

   // BAAAAAHAHHAAAHH!!!
   // here we switch the contexts!

   source = g_io_create_watch(client_io, flags);
   g_source_set_callback(source, (GSourceFunc) tcp_get_messages, 
			 client_info, NULL);
   g_source_attach(source, current_context->context);
   g_source_unref(source);

   client_info->context = current_context;
   client_info->gsource = source;
   current_context      = current_context->next;
   
   
   return TRUE;
}
  





/* *************************************************************************** *
 * set up the whole thing                                                      *
 * *************************************************************************** */

void tcp_setup()
{
   char         c_pure_ip_addr[16];
   GTcpSocket  *edk_socket;
   GTcpSocket  *ras_socket;


   // FUTURE: get ports & addresses from central_struct

   gint port_con = ED2K_CONNECTION_PORT;
   gint port_ras = RAS_CONNECTION_PORT;
   
   current_context = rome->contexts;

   // creation of server socket
   ras_socket = gnet_tcp_socket_server_new(port_ras);
   if (!ras_socket)
   {
	 fprintf(stdout, "tcp_connection_handler: couldnt create remote control server\naborting\n");
	 exit(-1);
   }

   // creation of client socket
   edk_socket = gnet_tcp_socket_server_new(port_con);
   if (!edk_socket)
   {
	 fprintf(stdout, "tcp_connection_handler: couldnt create ed2k server\n");
	 exit(-1);
   }


   gnet_tcp_socket_server_accept_async(edk_socket,
				       (GTcpSocketAcceptFunc) tcp_edonkey_connection, NULL);
   gnet_tcp_socket_server_accept_async(ras_socket, 
				       (GTcpSocketAcceptFunc) tcp_ras_connection, NULL);

   rome->running_addr    = gnet_tcp_socket_get_inetaddr(edk_socket);
   rome->outside_addr    = gnet_inetaddr_autodetect_internet_interface();
   rome->pure_ras_port   = RAS_CONNECTION_PORT;
   rome->hostname        = gnet_inetaddr_get_canonical_name(rome->outside_addr);

   if (!rome->pure_ipaddr || !rome->pure_conn_port)
      {
	 if (!rome->pure_ipaddr)
	    rome->pure_ipaddr     = inet_addr(gnet_inetaddr_get_canonical_name(rome->outside_addr));
	 if (!rome->pure_conn_port)
	    rome->pure_conn_port  = ED2K_CONNECTION_PORT;
	 inet_ntoc(rome->pure_ipaddr, c_pure_ip_addr);
	 g_print("tcp_connection_handler: using IP address %s:%d\n", 
		 c_pure_ip_addr, rome->pure_conn_port);
      }

   g_print("tcp_connection_handler: accepted on %s:%d\n", 
	   rome->hostname, gnet_inetaddr_get_port(rome->outside_addr));

   g_print("tcp_connection_handler: ras connection available on %s:%d\n", 
	   rome->hostname, rome->pure_ras_port);

}
