/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#include <gnet/gnet.h>

#include "structs.h"
#include "ed2k_data.h"
#include "ed2k_message_handler.h"
#include "settings.h"

extern struct central_struct *rome;




void 
remove_connection(struct UserInfo *client)
{
   // see ARCHITECTURE [2]

   // NONO: 
   // YES: 
   
   if (client->id)
      g_source_destroy(client->gsource);
   else
      g_source_remove(client->watch);

   g_io_channel_unref(client->iochannel);
   gnet_tcp_socket_delete(client->socket);
   
   
   if (client->id)
      {
	 ed2k_data_write_lock();
	 ed2k_data_user_delete(client);
	 ed2k_data_write_unlock();
      }


   g_free(client->inet_addr);
   //gnet_inetaddr_delete(client->addr);
   g_free(client);
}








gboolean
tcp_client_is_open(struct UserInfo *client)
{
   // return always true for now.

   GTcpSocket *contest;

   client->state = 10;

   contest = gnet_tcp_socket_connect(client->inet_addr, 
				     client->listen_port);
   
   if (!contest)
      {
	 /* #ifdef debug  */
	 /* 	 printf("\tchecking client ... client is firewalled\n"); */
	 /* #endif */

	 return FALSE;
      }
   else
      {
	 /* #ifdef debug  */
	 /* 	 printf("\tchecking client ... client is open\n"); */
	 /* #endif */

	 gnet_tcp_socket_delete(contest);
	 return TRUE;
      }
   

   g_assert_not_reached();
   return TRUE;
}






gboolean tcp_get_messages(GIOChannel *iochannel, GIOCondition condition, gpointer data)
{
   
   // always needed
   struct UserInfo *client_info;
   struct message_handle_struct message;

   // for the core functionality
   GError    *error = NULL;
   guint      read_bytes, 
              all_read_bytes,
              packet_size;
   gchar     *buffer;
   guchar     packet_peek[5];	 

   client_info = (struct UserInfo*) data;


   if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
      {
	 if (condition & G_IO_ERR)  
	    {
	       if (client_info->state == 10)
		  fprintf(stdout, "[ ERROR! ] client is being checked!\n");
	       else
		  {
		     fprintf(stdout, "[ ERROR! ] %s:%d - error or client disconnected. closing\n", 
			     client_info->inet_addr, client_info->port);
		     rome->debugdata.socket_error++;
		  }
	    }
	 else if (condition & G_IO_HUP)
	    {
	       fprintf(stdout, "[   S?   ] %s:%d - seems to be an unfriendly logout. hmm.\n", 
		       client_info->inet_addr, client_info->port);
	       rome->debugdata.unfriendly_hangup++;
	    }
	 else if (condition & G_IO_NVAL)
	    {
	       fprintf(stdout, "[  S??   ] %s:%d - I dont know what this is. but its ugly.\n", 
		       client_info->inet_addr, client_info->port);
	       rome->debugdata.dontknowwhat++;
	    }
	 goto __close_socket;
   }
   

   if (! (condition & G_IO_IN))
      goto __next;

	       
   /* check if socket is closed by client or receive data */

   all_read_bytes = 0;
   error = NULL;
   do
      {
	 g_io_channel_read_chars(iochannel, 
				 packet_peek, 
				 5, 
				 &read_bytes,
				 &error);
	 all_read_bytes += read_bytes;
      }
   while ((all_read_bytes < 5) && (error == NULL) && (read_bytes != 0));
	 

   // error checking orgy ... :-)
   if (read_bytes == 0) 
   {
      g_print("[ ERROR! ] from %s:%d: connection dropped. closing.\n", 
	      client_info->inet_addr, client_info->port);
      rome->stat_tcp.con_drops++;
      goto __close_socket;
   }
   else if (error) 
   {
	 printf("[ ERROR! ]   on %s:%d - closing socket\n", 
		client_info->inet_addr, client_info->port);
	 printf("[  ....  ] GError: %s\n", error->message);
	 g_clear_error(&error);
	 rome->debugdata.socket_error++;
	 goto __close_socket;
   }
   else if (packet_peek[0] != 227) 
   {
	 g_print("[ ERROR! ] from %s:%d : non edonkey packet - out of sync? closing.\n",
		 client_info->inet_addr, client_info->port);
	 goto __close_socket;
   }
	 
	 
   packet_size = *((int*)(packet_peek+1));
   buffer = g_new0(gchar, packet_size);
	 
	 
   all_read_bytes = 0;
   error = NULL;
   do
      {
	 g_io_channel_read_chars(iochannel,
				 buffer+all_read_bytes, 
				 packet_size - all_read_bytes,
				 &read_bytes,
				 &error);
	 all_read_bytes+=read_bytes;
      }
   while ((all_read_bytes != packet_size) && (read_bytes != 0) && (error==NULL));

   // error checking first
   if (error && ndl(2))
      {
	 fprintf(stdout, "[ ERROR! ] client: %s:%d: error reading message\n",
		 client_info->inet_addr, client_info->port);
	 fprintf(stdout, "[  ....  ] message: %s\n", error->message);
	 goto __close_socket;
      }
   else if (read_bytes == 0 && ndl(2)) 
      {
	 g_printerr("[ ERROR! ] client %s:%d connection dropped while recv'ing?\n",
		 client_info->inet_addr, client_info->port);
	 g_printerr("[  ....  ] closing socket.\n");
	 goto __close_socket;
      }

   message.size        = packet_size;
   message.buffer      = buffer;
   message.client_info = client_info;
   message.udp_address = NULL;
   message.udp_packet  = NULL;

   ed2k_handle_message(&message);
   
   
      __next:
   
   return TRUE;


      __close_socket:

   remove_connection(client_info);
   
   return FALSE;
}







void
tcp_context_listener(struct PollContext *context_struct)
{
   g_main_loop_run(context_struct->loop);

   printf("HUH? should not come back ... :-)\n");
}
