/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#include <gnet/gnet.h>
#include <stdio.h>


#include "structs.h"
#include "ed2k_message_gen.h"
#include "tcp_connections.h"
#include "ed2k_data.h"
#include "serverlist.h"
#include "settings.h"



/* ************************************************************************** */
/*                        for we all love globals                             */
/* ************************************************************************** */

extern struct central_struct *rome;
       
       struct DeleteList     *d_list;

              GMutex         *d_mutex;



/* ************************************************************************** */
/*                        the real timer functions                            */
/* ************************************************************************** */


gboolean
log_event_regular()
{
   time_t curr_time;


   if (cdl() > 3)
      printf("log_event_regular: called at offset %d\n",
	     (unsigned int) (time(NULL) - rome->start_time));


   // log:

   // time (timestamp)
   // users online
   // files online
   // bytes transmitted (tcp/udp/all)
   // bytes received (tcp/udp/all)
   // packets: udp/tcp count/bytes
   
   curr_time = time(NULL);

   fprintf(rome->log_files.regular_summary, 
	   // time users files_all files_diff files_bytes_all 
	   // traffic_in traf_in_tcp traf_in_udp (all in bytes+msgs)
	   // traf_out traf_out_tcp traf_out_udp (all in bytes+msgs)
	   // conn_reqs conn_drops errors
	   "%10d %5d %8d %8d %12d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %6d %6d %3d\n",
	   // time users
	   (unsigned int) curr_time, ed2k_data_get_users_count(), 
	   // files
	   ed2k_data_get_all_files_count(),
	   // files_different files_bytes
	   ed2k_data_get_different_files_count(), ed2k_data_get_files_byte_count(),

	   // traffic_in (tcp+udp) bytes
	   rome->stat_tcp.got_bytes+rome->stat_udp.got_bytes, 
	   // traffic_in (tcp+udp) messages
	   rome->stat_tcp.got_messages+rome->stat_udp.got_packets,
	   // traffic_in_tcp_bytes traf_in_tcp_messages
	   rome->stat_tcp.got_bytes, rome->stat_tcp.got_messages,
	   // traffic_in_udp_bytes traf_in_udp_messages
	   rome->stat_udp.got_bytes, rome->stat_udp.got_packets,

	   // traffic_out (tcp+udp) bytes
	   rome->stat_tcp.sent_bytes+rome->stat_udp.sent_bytes, 
	   // traffic_out (tcp+udp) messages
	   rome->stat_tcp.sent_messages+rome->stat_udp.sent_packets,
	   // traffic_out_tcp_bytes traf_out_tcp_messages
	   rome->stat_tcp.sent_bytes, rome->stat_tcp.sent_messages,
	   // traffic_out_udp_bytes traf_out_udp_messages
	   rome->stat_udp.sent_bytes, rome->stat_udp.sent_packets,

	   // conn_reqs conn_drops, 
	   rome->stat_tcp.con_reqs, rome->stat_tcp.con_drops,
	   // errors
	   rome->debugdata.socket_error + rome->debugdata.unfriendly_hangup +
	   rome->debugdata.dontknowwhat);

   // whooah!

   
   // 5 lines - TIME, IN bytes, IN messages, OUT bytes, OUT messages

   // TIME
   fprintf(rome->log_files.regular_details, "%d\n", (unsigned int) curr_time);

   // incoming tcp & udp (packets)
   fprintf(rome->log_files.regular_details, 
	   "%9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d\n", 
	   rome->stat_tcp.got_0x01, rome->stat_tcp.got_0x14, 
	   rome->stat_tcp.got_0x15, rome->stat_tcp.got_0x16, 
	   rome->stat_tcp.got_0x19, rome->stat_tcp.got_0x1c,
	   rome->stat_udp.got_0x96, rome->stat_udp.got_0x97, 
	   rome->stat_udp.got_0x98, rome->stat_udp.got_0x9a, 
	   rome->stat_udp.got_0x9c, rome->stat_udp.got_0xa0,
	   rome->stat_udp.got_0xa1, rome->stat_udp.got_0xa2, 
	   rome->stat_udp.got_0xa3, rome->stat_udp.got_0xa4);

   // incoming tcp & udp (bytes)
   fprintf(rome->log_files.regular_details, 
	   "%9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d\n", 
	   rome->stat_tcp.got_0x01_bytes, rome->stat_tcp.got_0x14_bytes, 
	   rome->stat_tcp.got_0x15_bytes, rome->stat_tcp.got_0x16_bytes, 
	   rome->stat_tcp.got_0x19_bytes, rome->stat_tcp.got_0x1c_bytes,
	   rome->stat_udp.got_0x96_bytes, rome->stat_udp.got_0x97_bytes, 
	   rome->stat_udp.got_0x98_bytes, rome->stat_udp.got_0x9a_bytes, 
	   rome->stat_udp.got_0x9c_bytes, rome->stat_udp.got_0xa0_bytes,
	   rome->stat_udp.got_0xa1_bytes, rome->stat_udp.got_0xa2_bytes, 
	   rome->stat_udp.got_0xa3_bytes, rome->stat_udp.got_0xa4_bytes);

   // outgoing tcp & udp (packets)
   fprintf(rome->log_files.regular_details, 
	   "%9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d\n",
	   rome->stat_tcp.sent_0x32, rome->stat_tcp.sent_0x33,
	   rome->stat_tcp.sent_0x34, rome->stat_tcp.sent_0x35,
	   rome->stat_tcp.sent_0x36, rome->stat_tcp.sent_0x38,
	   rome->stat_tcp.sent_0x40, rome->stat_tcp.sent_0x41,
	   rome->stat_tcp.sent_0x42,
	   rome->stat_udp.sent_0x96, rome->stat_udp.sent_0x97,
	   rome->stat_udp.sent_0x99, rome->stat_udp.sent_0x9b,
	   rome->stat_udp.sent_0x9e, rome->stat_udp.sent_0xa0, 
	   rome->stat_udp.sent_0xa1, rome->stat_udp.sent_0xa2, 
	   rome->stat_udp.sent_0xa3, rome->stat_udp.sent_0xa4);
   
   // outgoing tcp & udp (bytes)
   fprintf(rome->log_files.regular_details, 
	   "%9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d %9d\n",
	   rome->stat_tcp.sent_0x32_bytes, rome->stat_tcp.sent_0x33_bytes,
	   rome->stat_tcp.sent_0x34_bytes, rome->stat_tcp.sent_0x35_bytes,
	   rome->stat_tcp.sent_0x36_bytes, rome->stat_tcp.sent_0x38_bytes,
	   rome->stat_tcp.sent_0x40_bytes, rome->stat_tcp.sent_0x41_bytes,
	   rome->stat_tcp.sent_0x42_bytes,
	   rome->stat_udp.sent_0x96_bytes, rome->stat_udp.sent_0x97_bytes,
	   rome->stat_udp.sent_0x99_bytes, rome->stat_udp.sent_0x9b_bytes,
	   rome->stat_udp.sent_0x9e_bytes, rome->stat_udp.sent_0xa0_bytes, 
	   rome->stat_udp.sent_0xa1_bytes, rome->stat_udp.sent_0xa2_bytes, 
	   rome->stat_udp.sent_0xa3_bytes, rome->stat_udp.sent_0xa4_bytes);
   

   // last action!!
   if (rome->settings.new_logging_interval != rome->settings.logging_interval)
      {
	 rome->settings.logging_interval = rome->settings.new_logging_interval;
	 
	 g_timeout_add(rome->settings.new_logging_interval * 1000, 
		       (GSourceFunc) log_event_regular,
		       NULL);

	 return FALSE;
      }
   
   return TRUE;
}



guint
timed_ping_users(gpointer data)
{
   /*    printf("timed_ping_users: called at offset %d\n",  */
   /* 	  time(NULL) - rome->start_time); */
   return TRUE;
}


guint
timed_ping_other_servers(gpointer data)
{
   struct ServerList *servers;
   struct ServerList *parse;   
   int   i;
   guint curr_time;
   guint ip;

   server_lock();

   servers = server_get_all();

   if (!servers)
      return 1;

   
   if (cvl() > 3)
      printf("*timed event* timed_ping_other_servers: called at offset %d\n", 
	     (unsigned int) (time(NULL) - rome->start_time));


   if (cdl() > 3)
      printf("*timed event* checking ping timeouts\n");


   parse     = servers->prev;
   curr_time = time(NULL);

   while (parse != servers) 
      {
	 if (parse->challenge)
	    if (curr_time - parse->ping_start > 20)
	       {
		  parse->num_timeouts++;
		  if (parse->num_timeouts > 2)
		     {
			parse = parse->next;
			ip = parse->IP;
			server_delete(ip);
		     }
		  else
		     ed2k_send_udp_ping(parse->addr);
	       }
	 parse = parse->prev;
      }


   if (cdl() > 3)
      printf("*timed event* pinging servers\n");

   parse = servers;
   while (i < MAX_SERVERS_TO_PING && parse->challenge == 0 && parse != servers)
      {
	 ed2k_send_udp_ping(parse->addr);
	 i++;
      }

   
   server_unlock();
   
   return TRUE;
}




guint 
make_server_public(gpointer data)
{
   struct ServerList *serverlist;
   struct ServerList *server_start;

   if (! rome->settings.public)
      return 0;

   if (rome->settings.debuglevel > 3)
      printf("make_server_public: called at offset %d\n",
	     (unsigned int) (time(NULL) - rome->start_time));

   server_lock();
   
   serverlist   = server_get_all();
   server_start = serverlist;
   

   if (serverlist) 
      do
	 {
	    ed2k_send_udp_srv_announce(serverlist->addr);
	    ed2k_send_udp_ping((GInetAddr*) serverlist->addr);
	    serverlist->sent_announce = 1;
	    serverlist = serverlist->next;
	 } 
      while (serverlist != server_start);

   server_unlock();


   return FALSE;
}


guint
make_server_more_public(gpointer data)
{
   if (rome->settings.debuglevel > 3)
      printf("make_server_more_public: called at offset %d\n",
	     (unsigned int) (time(NULL) - rome->start_time));
   return FALSE;
}







/* ************************************************************************** */
/*                   one time functions for setting offsets                   */
/* ************************************************************************** */




guint
dispatch_server_timers(gpointer data)
{
   g_timeout_add(600000, (GSourceFunc)timed_ping_other_servers, NULL);
   return FALSE;
}


guint
dispatch_user_timers(gpointer data)
{
   g_timeout_add(120000, (GSourceFunc)timed_ping_users, NULL);
   return FALSE;
}


guint
dispatch_status_timers(gpointer data)
{
   g_timeout_add(300000, (GSourceFunc) log_event_regular,   NULL);
   return FALSE;
}




/* ************************************************************************** */
/*                           public deletion timers                           */
/* ************************************************************************** */



// wont probably be needed cause of readers locks ... :-)






/* ************************************************************************** */
/*                                set up                                      */
/* ************************************************************************** */


gint
timing_init()
{
   d_mutex = g_mutex_new();
   d_list  = NULL;
   

   return 1;
}


gint
timing_setup()
{
   g_timeout_add(30000, (GSourceFunc) dispatch_user_timers,    NULL);
   g_timeout_add(40000, (GSourceFunc) dispatch_server_timers,  NULL);
   g_timeout_add(50000, (GSourceFunc) dispatch_status_timers,  NULL);

   g_timeout_add( 5000, (GSourceFunc) make_server_public,      NULL);

   printf("set up timers.\n");


   return 1;
}
