/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#include <gnet/gnet.h>
#include <stdio.h>
#include <stdlib.h>

#include "structs.h"
#include "ed2k_message_handler.h"

#define MAX_UDP_PACKET_SIZE 4096



extern struct central_struct *rome;


int udp_message_handler(gpointer data)
{

   GUdpSocket   *server;
   GUdpPacket   *packet;
   GThreadPool  *processPool;

   guint n;
   
   struct message_handle_struct *message;
   gchar                        *buffer;
   
   processPool = rome->udp_pool;

   server = gnet_udp_socket_port_new(4665);
   if (!server)
      {
	 fprintf(stderr, "FATAL could not create UDP listener.\naborting.\n\n");
	 exit(-1);
      }

   g_assert(server);

   printf("udp_message_handler: waiting for packets (rome = %s)\n", rome->string);
   
   rome->udp_socket = server;
   
   n = 5;

   while (1)
      {
	 
	 if (n)
	    {
	       buffer  = g_new0(gchar, MAX_UDP_PACKET_SIZE);
	       packet  = gnet_udp_packet_receive_new(buffer, MAX_UDP_PACKET_SIZE);
	       message = g_new0(struct message_handle_struct, 1);
	    }

	 n = gnet_udp_socket_receive(server, packet);

	 if (!n)
	    {
	       continue;
	    }

	 g_assert(n);
	 
	 message->size        = n;
	 message->buffer      = packet->data;
	 message->udp_packet  = packet;
	 message->udp_address = (GInetAddr*) packet->addr;
	 g_thread_pool_push(processPool, message, NULL);
      }

   
   printf("[DEBUG] UDP handler ended.\n");

   g_assert_not_reached();
   
}




